package transactionCC

import (
	"gitlab.smartblocklab.com/smartcontracts/common/nums"
	"fmt"
	"errors"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"encoding/json"
	"reflect"
)

// TransactionRecord transaction struct which denotes transaction
// fields which are necessary for processing of
// transactions and will be used to present transaction in the ledger
type TransactionRecord struct {
	Key                        string                     `json:"key"`
	OperationID                string                     `json:"operation_id,omitempty"`
	ProcessType                ProcessType                `json:"process_type,omitempty"`
	OperationType              OperationType              `json:"operation_type,omitempty"`
	UserAccountInfo            *accountCC.UserAccountInfo `json:"user_account_info,omitempty"`
	DestinationUserAccountInfo *accountCC.UserAccountInfo `json:"destination_user_account_info,omitempty"`
	Amount                     nums.PlainCurrencyNumber   `json:"amount,omitempty"`
	HashCode                   string                     `json:"hash_code,omitempty"`
	DependencyKey              string                     `json:"dependency_key,omitempty"` // transaction dependency in a form of "txnType.currencyType.accountId"
	Timestamp                  Timestamp                  `json:"timestamp"`
	InOutType                  InputOutputType            `json:"input_output_type"`
}

func (this TransactionRecord) CheckBeforeAdd(manager chaincode.IManager) error {
	return nil
}

func (this TransactionRecord) CheckBeforeUpdate(updatedObject chaincode.ValidObject, manager chaincode.IManager) error {
	return errors.New("METHOD_DOES_NOT_IMPLEMENT")
}

func (this TransactionRecord) GetKey(manager chaincode.IManager) []string {
	key := fmt.Sprintf("%s_%s_%s_%s_%s", this.InOutType.String(), this.UserAccountInfo.UserID, this.Amount.CurrencyType, this.OperationType.String(), manager.GetAdapter().GetTxID())
	return []string{key}
}

func (this TransactionRecord) New() chaincode.ValidObject {
	return &TransactionRecord{}
}

func (this TransactionRecord) Validate() error {

	if (len(this.Key) == 0) {
		return errors.New("Key is empty")
	}

	if (len(this.DependencyKey) == 0) {
		return errors.New("DependencyKey is empty")
	}

	if (this.UserAccountInfo == nil) {
		return errors.New("UserAccountInfo is empty")
	}

	if len(this.UserAccountInfo.UserID) == 0 {
		return errors.New("UserID is empty")
	}

	if len(this.Amount.CurrencyType) == 0 {
		return errors.New("CurrencyType is empty")
	}

	if len(this.OperationID) == 0 {
		return errors.New("OperationID is empty")
	}

	if len(this.InOutType) == 0 {
		return errors.New("InOutType is empty")
	}

	if this.UserAccountInfo == nil {
		errMsg := fmt.Sprintf("UserAccountInfo can't be nil")
		return errors.New(errMsg)
	}

	err := this.UserAccountInfo.Validate()
	if err != nil {
		return err
	}

	// для входяших транзакций к пользователю на блокировку нужно делать проверку для поля DestinationUserAccountInfo
	// для остальных операций этого поля НЕТ
	if this.OperationType == LockAmount && this.InOutType == INPUT {

		if this.DestinationUserAccountInfo == nil {
			return errors.New("DestinationUserAccountInfo can't be nil")
		}

		err = this.DestinationUserAccountInfo.Validate()
		if err != nil {
			return errors.New("DestinationUserAccountInfo. " + err.Error())
		}
	}

	err = this.Amount.Validate()
	if err != nil {
		return err
	}

	return nil
}

func Unmarshal(data []byte) (*TransactionRecord, error) {
	var obj TransactionRecord
	err := json.Unmarshal(data, &obj)
	return &obj, err;
}

func Marshal(data *TransactionRecord) ([]byte, error) {
	return json.Marshal(data);
}

func Cast(validObject chaincode.ValidObject) (*TransactionRecord, error) {
	if validObject == nil {
		return nil, nil
	}

	if reflect.ValueOf(validObject).Kind() != reflect.Ptr {
		return nil, errors.New("check validObjectc is not reference")
	}

	casted, ok := validObject.(*TransactionRecord)
	if !ok {
		return nil, errors.New("Problem to get object")
	}

	return casted, nil;
}
